#include "stdio.h"
#include "file.h"

/*
 * typedefs
 */
typedef unsigned char   uchar;
typedef struct  Posn    Posn;
typedef struct  Dev Dev;
typedef struct  Etab    Etab;
typedef struct  Huff    Huff;

/*
 * structures
 */
struct  Huff
{
    FILE* fd;
    long    indexlen;
    long    grouplen;
    long    nspec;
    long    indexmod;
    long*   special;
    long*   tree;
    long*   ztree;
    long*   index;
};
struct  Etab
{
    int p[5];
    int type;
    int error;
    void*   hp;
};
struct  Posn
{
    char    bd[64];
    short   epp;
    short   mvno;
    short   mv[1];
};
struct  Dev
{
    long    h;
    int k1;
    int k2;
    int p1;
    int p2;
    int p3;
} dev;

/*
 * macros
 */
#define GONE    dev.k1

/*
 * constants
 */
enum
{
    MAXPLY  = 300,     /* was 200 */
    MAXGEN  = 100,
    BLACK   = 8,
    INIT    = 16,

    WPAWN   = 1,
    WKNIGHT,
    WBISHOP,
    WROOK,
    WQUEEN,
    WKING,

    BPAWN   = BLACK|WPAWN,
    BKNIGHT,
    BBISHOP,
    BROOK,
    BQUEEN,
    BKING,

    TMARK   = 0,
    TNORM,
    TEPENAB,
    TOOO,
    TOO,
    TNPRO,
    TBPRO,
    TRPRO,
    TQPRO,
    TENPAS,
    TNULL,

    U1  = 1<<0,
    U2  = 1<<1,
    L1  = 1<<2,
    L2  = 1<<3,
    D1  = 1<<4,
    D2  = 1<<5,
    R1  = 1<<6,
    R2  = 1<<7,

    ULEFT   = U1|L1,
    URIGHT  = U1|R1,
    DLEFT   = D1|L1,
    DRIGHT  = D1|R1,
    LEFT    = L1,
    RIGHT   = R1,
    UP  = U1,
    DOWN    = D1,
    U2R1    = U1|U2|R1,
    U1R2    = U1|R1|R2,
    D1R2    = D1|R1|R2,
    D2R1    = D1|D2|R1,
    D2L1    = D1|D2|L1,
    D1L2    = D1|L1|L2,
    U1L2    = U1|L1|L2,
    U2L1    = U1|U2|L1,
    RANK2   = D2,
    RANK7   = U2,

    INF = 999,
    NONE    = 990,

    AMARK   = 040000,
    EPMARK  = 020000,

    ENDFLAG = 0x800000,
};

/*
 * data
 */

/* cd */
extern  short*  lmp;
extern  short*  amp;
extern  short   ambuf[MAXPLY*2 + 10];
extern  int moveno;
extern  int eppos;
extern  int bkpos;
extern  int wkpos;
extern  char    board[64];
extern  short   mvlist[MAXPLY];
extern  int mvno;
extern  short   dir[64];
extern  short   attab[];
extern  short   color[];
extern  Posn    initp;
extern  char    ket[][4];
extern  Etab    etab[100];
extern  short   itab[1024];
extern  short   htab[64];
extern  short   gtab[1024];
extern  char    prm5[64];
/*
 * functions/subroutines
 */

/* libraries
extern  int open(char*, int);
extern  long    seek(int, long, int);
extern  long    read(int, void*, long);
extern  void    exit(int);   */

/* cm */
extern  int main(int, char*[]);
extern  void    conv2(void);
extern  void    conv4(void);
extern  void    unpackin(FILE*, long*, long, long);
extern  Huff*   huffinit(char*);
extern  int huffread(Huff*, long);
extern  char*   algout(int);
extern  void    xalgout(short*, int, char*);
extern  int mate(void);
extern  int match(int, int, int, int, int, int, int, int);
extern  int mat(int, int);

/* ch */
extern  int gethv(void);
extern  int wgethv(void);
extern  int endgame(void);
extern  int hscmp(int, int);
extern  int hwcmp(int, int);

/* cg */
extern  void    ginit(Posn*);
extern  void    getamv(short*);
extern  void    getmv(short*);
extern  void    move(int);
extern  int rmove(void);
extern  int retract(int);
extern  int check(void);
extern  void    wgen(void);
extern  int wcheck(int, int, int);
extern  void    wmove(int);
extern  void    wremove(void);
extern  int battack(int);
extern  void    bgen(void);
extern  int bcheck(int, int, int);
extern  void    bmove(int);
extern  void    bremove(void);
extern  int wattack(int);

/*
 * data -- initialized
 */
short   dir[] =
{
    U1|L1, U1|L2, U1   , U1   , U1   , U1   , U1|R2, U1|R1,
    U2|L1, U2|L2, U2   , U2   , U2   , U2   , U2|R2, U2|R1,
       L1,    L2,     0,     0,     0,     0,    R2,    R1,
       L1,    L2,     0,     0,     0,     0,    R2,    R1,
       L1,    L2,     0,     0,     0,     0,    R2,    R1,
       L1,    L2,     0,     0,     0,     0,    R2,    R1,
    D2|L1, D2|L2, D2   , D2   , D2   , D2   , D2|R2, D2|R1,
    D1|L1, D1|L2, D1   , D1   , D1   , D1   , D1|R2, D1|R1,
};
short   attab[] =
{
    U2R1, -15,
    U1R2, -6,
    D1R2, 10,
    D2R1, 17,
    D2L1, 15,
    D1L2, 6,
    U1L2, -10,
    U2L1, -17,

    ULEFT, -9,
    URIGHT, -7,
    DLEFT, 7,
    DRIGHT, 9,

    UP, -8,
    LEFT, -1,
    RIGHT, 1,
    DOWN, 8,
};
Posn    initp =
{
    INIT+BROOK,
    INIT+BKNIGHT,
    INIT+BBISHOP,
    INIT+BQUEEN,
    INIT+BKING,
    INIT+BBISHOP,
    INIT+BKNIGHT,
    INIT+BROOK,
    INIT+BPAWN,
    INIT+BPAWN,
    INIT+BPAWN,
    INIT+BPAWN,
    INIT+BPAWN,
    INIT+BPAWN,
    INIT+BPAWN,
    INIT+BPAWN,
    0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,
    INIT+WPAWN,
    INIT+WPAWN,
    INIT+WPAWN,
    INIT+WPAWN,
    INIT+WPAWN,
    INIT+WPAWN,
    INIT+WPAWN,
    INIT+WPAWN,
    INIT+WROOK,
    INIT+WKNIGHT,
    INIT+WBISHOP,
    INIT+WQUEEN,
    INIT+WKING,
    INIT+WBISHOP,
    INIT+WKNIGHT,
    INIT+WROOK,
    0,  /* eppos */
    0,  /* moveno */
    0,  /* moves */
};

short   color[] =
{
     0,
     1,  1,  1,  1,  1,  1,  1,
     0,
    -1, -1, -1, -1, -1, -1, -1,
     0,
     1,  1,  1,  1,  1,  1,  1,
     0,
    -1, -1, -1, -1, -1, -1, -1,
};
char    ket[][4] =
{
/*
 * white, white, black
 */

    WKNIGHT,    WKNIGHT,    BBISHOP,    2,
    WKNIGHT,    WKNIGHT,    BKNIGHT,    2,
    WKNIGHT,    WKNIGHT,    BQUEEN,     2,
    WKNIGHT,    WKNIGHT,    BROOK,      2,

    WQUEEN,     BQUEEN,     BROOK,      2,
    WQUEEN,     BROOK,      BBISHOP,    2,
    WQUEEN,     BROOK,      BKNIGHT,    2,
    WQUEEN,     BROOK,      BROOK,      2,

    WROOK,      BBISHOP,    BBISHOP,    2,
    WROOK,      BBISHOP,    BKNIGHT,    2,
    WROOK,      BKNIGHT,    BKNIGHT,    2,
    WROOK,      BQUEEN,     BBISHOP,    2,
    WROOK,      BQUEEN,     BKNIGHT,    2,
    WROOK,      BROOK,      BBISHOP,    2,
    WROOK,      BROOK,      BKNIGHT,    2,

    WQUEEN,     WROOK,      BBISHOP,    2,
    WROOK,      WBISHOP,    BBISHOP,    2,
    WROOK,      WKNIGHT,    BBISHOP,    2,
    WROOK,      WROOK,      BBISHOP,    2,


    WQUEEN,     WBISHOP,    BBISHOP,    2,
    WQUEEN,     WBISHOP,    BKNIGHT,    2,
    WQUEEN,     WKNIGHT,    BBISHOP,    2,
    WQUEEN,     WROOK,      BQUEEN,     2,
    WROOK,      WBISHOP,    BBISHOP,    2,
    WROOK,      WBISHOP,    BKNIGHT,    2,
    WROOK,      WBISHOP,    BQUEEN,     2,
    WROOK,      WKNIGHT,    BBISHOP,    2,
    WROOK,      WKNIGHT,    BQUEEN,     2,
    WROOK,      WROOK,      BQUEEN,     2,
    WBISHOP,    WBISHOP,    BBISHOP,    2,
    WBISHOP,    WBISHOP,    BKNIGHT,    2,
    WBISHOP,    WKNIGHT,    BBISHOP,    2,
    WBISHOP,    WKNIGHT,    BKNIGHT,    2,
    WKNIGHT,    WKNIGHT,    BBISHOP,    2,

/*
 * a pawn
 */

    WPAWN,      WBISHOP,    BBISHOP,    4,
    WPAWN,      WBISHOP,    BKNIGHT,    4,
    WPAWN,      WKNIGHT,    BBISHOP,    4,
    WPAWN,      WROOK,      BQUEEN,     4,


/*
 * a pawn
 */
    WPAWN,      WROOK,      BBISHOP,    4,
    BPAWN,      WKNIGHT,    WKNIGHT,    4,
    BPAWN,      WQUEEN,     BROOK,      4,
    BPAWN,      WROOK,      BBISHOP,    4,
    BPAWN,      WROOK,      WKNIGHT,    4,

    0
};
short   itab[1024] =
{
      462,  462,    0,    1,    2,    3,    4,    5,
      462,  462,    6,    7,    8,    9,   10,   11,
      462,  462,   12,   13,   14,   15,   16,   17,
      462,  462,  462,   18,   19,   20,   21,   22,
      462,  462,  462,  462,   23,   24,   25,   26,
      462,  462,  462,  462,  462,   27,   28,   29,
      462,  462,  462,  462,  462,  462,   30,   31,
      462,  462,  462,  462,  462,  462,  462,   32,
      462,  462,  462,   33,   34,   35,   36,   37,
      462,  462,  462,   38,   39,   40,   41,   42,
       43,   44,   45,   46,   47,   48,   49,   50,
       51,   52,   53,   54,   55,   56,   57,   58,
       59,   60,   61,   62,   63,   64,   65,   66,
       67,   68,   69,   70,   71,   72,   73,   74,
       75,   76,   77,   78,   79,   80,   81,   82,
       83,   84,   85,   86,   87,   88,   89,   90,
       91,  462,  462,  462,   92,   93,   94,   95,
       96,  462,  462,  462,   97,   98,   99,  100,
      101,  102,  103,  104,  105,  106,  107,  108,
      109,  110,  111,  112,  113,  114,  115,  116,
      117,  118,  119,  120,  121,  122,  123,  124,
      125,  126,  127,  128,  129,  130,  131,  132,
      133,  134,  135,  136,  137,  138,  139,  140,
      141,  142,  143,  144,  145,  146,  147,  148,
      149,  150,  462,  462,  462,  151,  152,  153,
      154,  155,  462,  462,  462,  156,  157,  158,
      159,  160,  161,  162,  163,  164,  165,  166,
      167,  168,  169,  170,  171,  172,  173,  174,
      175,  176,  177,  178,  179,  180,  181,  182,
      183,  184,  185,  186,  187,  188,  189,  190,
      191,  192,  193,  194,  195,  196,  197,  198,
      199,  200,  201,  202,  203,  204,  205,  206,
      462,  462,  462,  207,  208,  209,  210,  211,
      462,  462,  462,  212,  213,  214,  215,  216,
      462,  462,  462,  217,  218,  219,  220,  221,
      462,  462,  462,  222,  223,  224,  225,  226,
      462,  462,  462,  462,  227,  228,  229,  230,
      462,  462,  462,  462,  462,  231,  232,  233,
      462,  462,  462,  462,  462,  462,  234,  235,
      462,  462,  462,  462,  462,  462,  462,  236,
      237,  462,  462,  462,  238,  239,  240,  241,
      242,  462,  462,  462,  243,  244,  245,  246,
      247,  462,  462,  462,  248,  249,  250,  251,
      252,  253,  254,  255,  256,  257,  258,  259,
      260,  261,  262,  263,  264,  265,  266,  267,
      268,  269,  270,  271,  272,  273,  274,  275,
      276,  277,  278,  279,  280,  281,  282,  283,
      284,  285,  286,  287,  288,  289,  290,  291,
      292,  293,  462,  462,  462,  294,  295,  296,
      297,  298,  462,  462,  462,  299,  300,  301,
      302,  303,  462,  462,  462,  304,  305,  306,
      307,  308,  309,  310,  311,  312,  313,  314,
      315,  316,  317,  318,  319,  320,  321,  322,
      323,  324,  325,  326,  327,  328,  329,  330,
      331,  332,  333,  334,  335,  336,  337,  338,
      339,  340,  341,  342,  343,  344,  345,  346,
      347,  348,  349,  350,  351,  352,  353,  354,
      462,  462,  462,  462,  355,  356,  357,  358,
      462,  462,  462,  462,  359,  360,  361,  362,
      462,  462,  462,  462,  363,  364,  365,  366,
      462,  462,  462,  462,  367,  368,  369,  370,
      462,  462,  462,  462,  462,  371,  372,  373,
      462,  462,  462,  462,  462,  462,  374,  375,
      462,  462,  462,  462,  462,  462,  462,  376,
      377,  378,  379,  380,  381,  382,  383,  384,
      385,  386,  462,  462,  462,  387,  388,  389,
      390,  391,  462,  462,  462,  392,  393,  394,
      395,  396,  462,  462,  462,  397,  398,  399,
      400,  401,  402,  403,  404,  405,  406,  407,
      408,  409,  410,  411,  412,  413,  414,  415,
      416,  417,  418,  419,  420,  421,  422,  423,
      424,  425,  426,  427,  428,  429,  430,  431,
      432,  433,  434,  435,  436,  437,  438,  439,
      462,  440,  441,  442,  443,  444,  445,  446,
      462,  462,  462,  462,  462,  447,  448,  449,
      462,  462,  462,  462,  462,  450,  451,  452,
      462,  462,  462,  462,  462,  453,  454,  455,
      462,  462,  462,  462,  462,  456,  457,  458,
      462,  462,  462,  462,  462,  462,  459,  460,
      462,  462,  462,  462,  462,  462,  462,  461,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
      462,  462,  462,  462,  462,  462,  462,  462,
};
short   gtab[1024] =
{
    0000,0001,0002,0003,0004,0005,0006,0007,
    0010,0011,0012,0013,0014,0015,0016,0017,
    0020,0021,0022,0023,0024,0025,0026,0027,
    0030,0031,0032,0033,0034,0035,0036,0037,
    0040,0041,0042,0043,0044,0045,0046,0047,
    0050,0051,0052,0053,0054,0055,0056,0057,
    0060,0061,0062,0063,0064,0065,0066,0067,
    0070,0071,0072,0073,0074,0075,0076,0077,
    0100,0001,0002,0003,0004,0005,0006,0007,
    0201,0111,0012,0013,0014,0015,0016,0017,
    0202,0212,0122,0023,0024,0025,0026,0027,
    0203,0213,0223,0133,0034,0035,0036,0037,
    0204,0214,0224,0234,0144,0045,0046,0047,
    0205,0215,0225,0235,0245,0155,0056,0057,
    0206,0216,0226,0236,0246,0256,0166,0067,
    0207,0217,0227,0237,0247,0257,0267,0177,
    0200,0210,0220,0230,0240,0250,0260,0270,
    0201,0211,0221,0231,0241,0251,0261,0271,
    0202,0212,0222,0232,0242,0252,0262,0272,
    0203,0213,0223,0233,0243,0253,0263,0273,
    0204,0214,0224,0234,0244,0254,0264,0274,
    0205,0215,0225,0235,0245,0255,0265,0275,
    0206,0216,0226,0236,0246,0256,0266,0276,
    0207,0217,0227,0237,0247,0257,0267,0277,
    0300,0210,0220,0230,0240,0250,0260,0270,
    0201,0311,0221,0231,0241,0251,0261,0271,
    0202,0212,0322,0232,0242,0252,0262,0272,
    0203,0213,0223,0333,0243,0253,0263,0273,
    0204,0214,0224,0234,0344,0254,0264,0274,
    0205,0215,0225,0235,0245,0355,0265,0275,
    0206,0216,0226,0236,0246,0256,0366,0276,
    0207,0217,0227,0237,0247,0257,0267,0377,
    0007,0006,0005,0004,0003,0002,0001,0000,
    0017,0016,0015,0014,0013,0012,0011,0010,
    0027,0026,0025,0024,0023,0022,0021,0020,
    0037,0036,0035,0034,0033,0032,0031,0030,
    0047,0046,0045,0044,0043,0042,0041,0040,
    0057,0056,0055,0054,0053,0052,0051,0050,
    0067,0066,0065,0064,0063,0062,0061,0060,
    0077,0076,0075,0074,0073,0072,0071,0070,
    0007,0006,0005,0004,0003,0002,0001,0100,
    0017,0016,0015,0014,0013,0012,0111,0201,
    0027,0026,0025,0024,0023,0122,0212,0202,
    0037,0036,0035,0034,0133,0223,0213,0203,
    0047,0046,0045,0144,0234,0224,0214,0204,
    0057,0056,0155,0245,0235,0225,0215,0205,
    0067,0166,0256,0246,0236,0226,0216,0206,
    0177,0267,0257,0247,0237,0227,0217,0207,
    0270,0260,0250,0240,0230,0220,0210,0200,
    0271,0261,0251,0241,0231,0221,0211,0201,
    0272,0262,0252,0242,0232,0222,0212,0202,
    0273,0263,0253,0243,0233,0223,0213,0203,
    0274,0264,0254,0244,0234,0224,0214,0204,
    0275,0265,0255,0245,0235,0225,0215,0205,
    0276,0266,0256,0246,0236,0226,0216,0206,
    0277,0267,0257,0247,0237,0227,0217,0207,
    0270,0260,0250,0240,0230,0220,0210,0300,
    0271,0261,0251,0241,0231,0221,0311,0201,
    0272,0262,0252,0242,0232,0322,0212,0202,
    0273,0263,0253,0243,0333,0223,0213,0203,
    0274,0264,0254,0344,0234,0224,0214,0204,
    0275,0265,0355,0245,0235,0225,0215,0205,
    0276,0366,0256,0246,0236,0226,0216,0206,
    0377,0267,0257,0247,0237,0227,0217,0207,
    0070,0071,0072,0073,0074,0075,0076,0077,
    0060,0061,0062,0063,0064,0065,0066,0067,
    0050,0051,0052,0053,0054,0055,0056,0057,
    0040,0041,0042,0043,0044,0045,0046,0047,
    0030,0031,0032,0033,0034,0035,0036,0037,
    0020,0021,0022,0023,0024,0025,0026,0027,
    0010,0011,0012,0013,0014,0015,0016,0017,
    0000,0001,0002,0003,0004,0005,0006,0007,
    0207,0217,0227,0237,0247,0257,0267,0177,
    0206,0216,0226,0236,0246,0256,0166,0067,
    0205,0215,0225,0235,0245,0155,0056,0057,
    0204,0214,0224,0234,0144,0045,0046,0047,
    0203,0213,0223,0133,0034,0035,0036,0037,
    0202,0212,0122,0023,0024,0025,0026,0027,
    0201,0111,0012,0013,0014,0015,0016,0017,
    0100,0001,0002,0003,0004,0005,0006,0007,
    0207,0217,0227,0237,0247,0257,0267,0277,
    0206,0216,0226,0236,0246,0256,0266,0276,
    0205,0215,0225,0235,0245,0255,0265,0275,
    0204,0214,0224,0234,0244,0254,0264,0274,
    0203,0213,0223,0233,0243,0253,0263,0273,
    0202,0212,0222,0232,0242,0252,0262,0272,
    0201,0211,0221,0231,0241,0251,0261,0271,
    0200,0210,0220,0230,0240,0250,0260,0270,
    0207,0217,0227,0237,0247,0257,0267,0377,
    0206,0216,0226,0236,0246,0256,0366,0276,
    0205,0215,0225,0235,0245,0355,0265,0275,
    0204,0214,0224,0234,0344,0254,0264,0274,
    0203,0213,0223,0333,0243,0253,0263,0273,
    0202,0212,0322,0232,0242,0252,0262,0272,
    0201,0311,0221,0231,0241,0251,0261,0271,
    0300,0210,0220,0230,0240,0250,0260,0270,
    0077,0076,0075,0074,0073,0072,0071,0070,
    0067,0066,0065,0064,0063,0062,0061,0060,
    0057,0056,0055,0054,0053,0052,0051,0050,
    0047,0046,0045,0044,0043,0042,0041,0040,
    0037,0036,0035,0034,0033,0032,0031,0030,
    0027,0026,0025,0024,0023,0022,0021,0020,
    0017,0016,0015,0014,0013,0012,0011,0010,
    0007,0006,0005,0004,0003,0002,0001,0000,
    0177,0267,0257,0247,0237,0227,0217,0207,
    0067,0166,0256,0246,0236,0226,0216,0206,
    0057,0056,0155,0245,0235,0225,0215,0205,
    0047,0046,0045,0144,0234,0224,0214,0204,
    0037,0036,0035,0034,0133,0223,0213,0203,
    0027,0026,0025,0024,0023,0122,0212,0202,
    0017,0016,0015,0014,0013,0012,0111,0201,
    0007,0006,0005,0004,0003,0002,0001,0100,
    0277,0267,0257,0247,0237,0227,0217,0207,
    0276,0266,0256,0246,0236,0226,0216,0206,
    0275,0265,0255,0245,0235,0225,0215,0205,
    0274,0264,0254,0244,0234,0224,0214,0204,
    0273,0263,0253,0243,0233,0223,0213,0203,
    0272,0262,0252,0242,0232,0222,0212,0202,
    0271,0261,0251,0241,0231,0221,0211,0201,
    0270,0260,0250,0240,0230,0220,0210,0200,
    0377,0267,0257,0247,0237,0227,0217,0207,
    0276,0366,0256,0246,0236,0226,0216,0206,
    0275,0265,0355,0245,0235,0225,0215,0205,
    0274,0264,0254,0344,0234,0224,0214,0204,
    0273,0263,0253,0243,0333,0223,0213,0203,
    0272,0262,0252,0242,0232,0322,0212,0202,
    0271,0261,0251,0241,0231,0221,0311,0201,
    0270,0260,0250,0240,0230,0220,0210,0300,
};
short   htab[64] =
{
    0100,0001,0002,0003,0023,0022,0021,0120,
    0201,0104,0005,0006,0026,0025,0124,0221,
    0202,0205,0107,0010,0030,0127,0225,0222,
    0203,0206,0210,0111,0131,0230,0226,0223,
    0243,0246,0250,0151,0171,0270,0266,0263,
    0242,0245,0147,0050,0070,0167,0265,0262,
    0241,0144,0045,0046,0066,0065,0164,0261,
    0140,0041,0042,0043,0063,0062,0061,0160,
};
char    prm5[64] =
{
     7, 6, 5, 4, 3, 2, 1, 0,
    15,14,13,12,11,10, 9, 8,
    23,22,21,20,19,18,17,16,
    31,30,29,28,27,26,25,24,
    39,38,37,36,35,34,33,32,
    47,46,45,44,43,42,41,40,
    55,54,53,52,51,50,49,48,
    63,62,61,60,59,58,57,56,
};

int pit = 0;
int badflag = 1;
int badbad = 0;
/*
 * data -- uninitialized
 */
Etab    etab[100];
short*  lmp;
short*  amp;
short   ambuf[MAXPLY*2 + 10];
int moveno;
int eppos;
int bkpos;
int wkpos;
char    board[64];
short   mvlist[MAXPLY];
int mvno;
int optcnt;
/*
 * chess move generator
 */
void
ginit(Posn *posn)
{
    int i;

    for(i=0; i<64; i++) {
        board[i] = posn->bd[i];
        if((board[i]&017) == WKING)
            wkpos = i;
        if((board[i]&017) == BKING)
            bkpos = i;
    }
    eppos = posn->epp;
    moveno = posn->mvno;
    mvno = 0;
    mvlist[0] = 0;
    amp = ambuf;
    *amp++ = 0;
    for(i=0; posn->mv[i]; i++)
        move(posn->mv[i]);
}

void
getamv(short *mv)
{

    lmp = mv;
    if(moveno & 1)
        bgen();
    else
        wgen();
    *lmp = 0;
}

void
getmv(short *mv)
{
    int i, f;

    lmp = mv;
    if(moveno & 1)
        bgen();
    else
        wgen();
    *lmp = 0;
    lmp = mv;
    for(i=0; mv[i]; i++) {
        if(moveno & 1) {
            bmove(mv[i]);
            f = wattack(bkpos);
            bremove();
        } else {
            wmove(mv[i]);
            f = battack(wkpos);
            wremove();
        }
        if(f)
            *lmp++ = mv[i];
    }
    *lmp = 0;
}

void
move(int mv)
{

    if(moveno & 1)
        bmove(mv);
    else
        wmove(mv);
    moveno++;
    mvlist[mvno++] = mv;
}

int
rmove(void)
{

    if(mvno <= 0)
        return 0;
    moveno--;
    if(moveno & 1)
        bremove();
    else
        wremove();
    return mvlist[--mvno];
}

int
retract(int m)
{
    int p;

    if(m == 0)
        return 1;
    p = board[(m>>6) & 077] & 7;
    if(p == WPAWN)
        return 1;
    if(p == WKING) {
        p = (m>>12) & 017;
        if(p == TOO || p == TOOO)
            return 1;
    }
    p = board[m & 077] & 7;
    if(p != 0)
        return 1;
    return 0;
}

int
check(void)
{

    if(moveno & 1)
        return !wattack(bkpos);
    return !battack(wkpos);
}

void
wgen(void)
{
    short *p;
    int i, j, k, d, p0, p1;

    for(i=0; i<64; i++)
    if(board[i])
    switch(board[i]) {

    case INIT|WPAWN:
    case WPAWN:
        d = dir[i];
        if((d&ULEFT) == 0) {
            if(color[board[i-9]] < 0)
            if((d&RANK7) != 0) {
                wcheck(i, i-9, TNPRO);
                wcheck(i, i-9, TBPRO);
                wcheck(i, i-9, TRPRO);
                wcheck(i, i-9, TQPRO);
            } else
                wcheck(i, i-9, TNORM);
            if(eppos == i-1)
                wcheck(i, i-9, TENPAS);
        }
        if((d&URIGHT) == 0) {
            if(color[board[i-7]] < 0)
            if((d&RANK7) != 0) {
                wcheck(i, i-7, TNPRO);
                wcheck(i, i-7, TBPRO);
                wcheck(i, i-7, TRPRO);
                wcheck(i, i-7, TQPRO);
            } else
                wcheck(i, i-7, TNORM);
            if(eppos == i+1)
                wcheck(i, i-7, TENPAS);
        }
        if(board[i-8] == 0) {
            if((d&RANK7) != 0) {
                wcheck(i, i-8, TNPRO);
                wcheck(i, i-8, TBPRO);
                wcheck(i, i-8, TRPRO);
                wcheck(i, i-8, TQPRO);
            } else
                wcheck(i, i-8, TNORM);
            if(d & RANK2)
            if(board[i-16] == 0)
                wcheck(i, i-16, TEPENAB);
        }
        continue;

    case INIT|WKNIGHT:
    case WKNIGHT:
        p = attab;
        goto jump;

    case INIT|WKING:
        if(board[56] == (INIT|WROOK))
        if(board[57] == 0)
        if(board[58] == 0)
        if(board[59] == 0)
        if(battack(58))
        if(battack(59))
        if(battack(60))
        if(i == 60)
            wcheck(60, 58, TOOO);
        if(board[63] == (INIT|WROOK))
        if(board[62] == 0)
        if(board[61] == 0)
        if(battack(62))
        if(battack(61))
        if(battack(60))
        if(i == 60)
            wcheck(60, 62, TOO);

    case WKING:
        p = attab+16;

    jump:
        d = dir[i];
        for(j=8; j--; p+=2)
            if((d&p[0]) == 0)
                wcheck(i, i+p[1], TNORM);
        continue;

    case INIT|WBISHOP:
    case WBISHOP:
        p = attab+16;
        j = 4;
        goto slide;

    case INIT|WROOK:
    case WROOK:
        p = attab+24;
        j = 4;
        goto slide;

    case INIT|WQUEEN:
    case WQUEEN:
        p = attab+16;
        j = 8;

    slide:
        while(j--) {
            p0 = *p++;
            p1 = *p++;
            k = i;
            while((dir[k]&p0) == 0) {
                k += p1;
                if(wcheck(i, k, TNORM))
                    break;
            }
        }
        continue;
    }
}

int
wcheck(int from, int to, int typ)
{
    int b;

    b = board[to];
    if(color[b] > 0)
        return 1;
    *lmp++ = (typ<<12) | (from<<6) | to;
    return b;
}

void
wmove(int mov)
{
    int i, from, to;

    from = (mov>>6) & 077;
    to = (mov>>0) & 077;
    eppos = 0;
    *amp++ = AMARK | (from<<5) | board[from];
    if(from == 0 && to == 0)
        return;
    *amp++ = (to<<5) | board[to];
    i = board[from] & 017;
    board[to] = i;
    if(i == WKING)
        wkpos = to;
    board[from] = 0;
    switch((mov>>12)&017) {

    case TNPRO:
        board[to] = WKNIGHT;
        break;

    case TBPRO:
        board[to] = WBISHOP;
        break;

    case TRPRO:
        board[to] = WROOK;
        break;

    case TQPRO:
        board[to] = WQUEEN;
        break;

    case TOO:
        from = to+1;
        to--;
        goto mv;

    case TOOO:
        from = to-2;
        to++;

    mv:
        *amp++ = (from<<5) | board[from];
        *amp++ = (to<<5) | board[to];
        board[to] = board[from] & 017;
        board[from] = 0;
        break;

    case TENPAS:
        to += 8;
        *amp++ = (to<<5) | board[to];
        board[to] = 0;
        break;

    case TEPENAB:
        amp[-1] |= EPMARK;
        eppos = to;
        break;
    }
}

void
wremove(void)
{
    int i, pi;

    do {
        amp--;
        i = (*amp>>5) & 077;
        pi = *amp & 037;
        board[i] = pi;
        if((pi&017) == WKING)
            wkpos = i;
    } while((*amp&AMARK) == 0);
    eppos = 0;
    if(amp[-1] & EPMARK)
        eppos = (amp[-1]>>5) & 077;
}

int
battack(int pos)
{
    short *p;
    int i, j, p0, p1, b, d;

    p = attab;
    d = dir[pos];

    /* attack by knight */
    for(i=8; i--; p+=2) {
        if((d & p[0]) == 0)
        if((board[pos+p[1]]&017) == BKNIGHT)
            return 0;
    }

    /* attack along diagonal */
    for(i=4; i--;) {
        p0 = *p++;
        p1 = *p++;
        j = pos;
        while((dir[j]&p0) == 0) {
            j += p1;
            b = board[j] & 017;
            if(b)
            if(b == BQUEEN || b == BBISHOP)
                return 0;
            else
                break;
        }
    }

    /* attack at right angles */
    for(i=4; i--;) {
        p0 = *p++;
        p1 = *p++;
        j = pos;
        while((dir[j]&p0) == 0) {
            j += p1;
            b = board[j] & 017;
            if(b)
            if(b == BQUEEN || b == BROOK)
                return 0;
            else
                break;
        }
    }

    /* attack by king */
    p -= 16;
    for(i=8; i--; p+=2)
        if((d & p[0]) == 0)
        if(board[pos+p[1]] == BKING)
            return 0;

    /* attack by pawns */
    if((d & ULEFT) == 0)
    if((board[pos-9]&017) == BPAWN)
        return 0;
    if((d & URIGHT) == 0)
    if((board[pos-7]&017) == BPAWN)
        return 0;

    return 1;
}

void
bgen(void)
{
    short *p;
    int i, j, k, d, p0, p1;

    for(i=0; i<64; i++)
    if(board[i])
    switch(board[i]) {

    case INIT|BPAWN:
    case BPAWN:
        d = dir[i];
        if((d&DLEFT) == 0) {
            if(color[board[i+7]] > 0)
            if((d&RANK2) != 0) {
                bcheck(i, i+7, TNPRO);
                bcheck(i, i+7, TBPRO);
                bcheck(i, i+7, TRPRO);
                bcheck(i, i+7, TQPRO);
            } else
                bcheck(i, i+7, TNORM);
            if(eppos == i-1)
                bcheck(i, i+7, TENPAS);
        }
        if((d&DRIGHT) == 0) {
            if(color[board[i+9]] > 0)
            if((d&RANK2) != 0) {
                bcheck(i, i+9, TNPRO);
                bcheck(i, i+9, TBPRO);
                bcheck(i, i+9, TRPRO);
                bcheck(i, i+9, TQPRO);
            } else
                bcheck(i, i+9, TNORM);
            if(eppos == i+1)
                bcheck(i, i+9, TENPAS);
        }
        if(board[i+8] == 0) {
            if((d&RANK2) != 0) {
                bcheck(i, i+8, TNPRO);
                bcheck(i, i+8, TBPRO);
                bcheck(i, i+8, TRPRO);
                bcheck(i, i+8, TQPRO);
            } else
                bcheck(i, i+8, TNORM);
            if(d & RANK7)
            if(board[i+16] == 0)
                bcheck(i, i+16, TEPENAB);
        }
        continue;

    case INIT|BKNIGHT:
    case BKNIGHT:
        p = attab;
        goto jump;

    case INIT|BKING:
        if(board[0] == (INIT|BROOK))
        if(board[1] == 0)
        if(board[2] == 0)
        if(board[3] == 0)
        if(wattack(2))
        if(wattack(3))
        if(wattack(4))
        if(i == 4)
            bcheck(4, 2, TOOO);
        if(board[7] == (INIT|BROOK))
        if(board[6] == 0)
        if(board[5] == 0)
        if(wattack(6))
        if(wattack(5))
        if(wattack(4))
        if(i == 4)
            bcheck(4, 6, TOO);

    case BKING:
        p = attab+16;

    jump:
        d = dir[i];
        for(j=8; j--; p+=2)
            if((d&p[0]) == 0)
                bcheck(i, i+p[1], TNORM);
        continue;

    case INIT|BBISHOP:
    case BBISHOP:
        p = attab+16;
        j = 4;
        goto slide;

    case INIT|BROOK:
    case BROOK:
        p = attab+24;
        j = 4;
        goto slide;

    case INIT|BQUEEN:
    case BQUEEN:
        p = attab+16;
        j = 8;

    slide:
        while(j--) {
            p0 = *p++;
            p1 = *p++;
            k = i;
            while((dir[k]&p0) == 0) {
                k += p1;
                if(bcheck(i, k, TNORM))
                    break;
            }
        }
        continue;
    }
}

int
bcheck(int from, int to, int typ)
{
    int b;

    b = board[to];
    if(color[b] < 0)
        return 1;
    *lmp++ = (typ<<12) | (from<<6) | to;
    return b;
}

void
bmove(int mov)
{
    int i, from, to;

    from = (mov>>6) & 077;
    to = (mov>>0) & 077;
    eppos = 0;
    *amp++ = AMARK | (from<<5) | board[from];
    if(from == 0 && to == 0)
        return;
    *amp++ = (to<<5) | board[to];
    i = board[from] & 017;
    board[to] = i;
    if(i == BKING)
        bkpos = to;
    board[from] = 0;
    switch((mov>>12)&017) {

    case TNPRO:
        board[to] = BKNIGHT;
        break;

    case TBPRO:
        board[to] = BBISHOP;
        break;

    case TRPRO:
        board[to] = BROOK;
        break;

    case TQPRO:
        board[to] = BQUEEN;
        break;

    case TOO:
        from = to+1;
        to--;
        goto mv;

    case TOOO:
        from = to-2;
        to++;

    mv:
        *amp++ = (from<<5) | board[from];
        *amp++ = (to<<5) | board[to];
        board[to] = board[from] & 017;
        board[from] = 0;
        break;

    case TENPAS:
        to -= 8;
        *amp++ = (to<<5) | board[to];
        board[to] = 0;
        break;

    case TEPENAB:
        amp[-1] |= EPMARK;
        eppos = to;
        break;
    }
}

void
bremove(void)
{
    int i, pi;

    do {
        amp--;
        i = (*amp>>5) & 077;
        pi = *amp & 037;
        board[i] = pi;
        if((pi&017) == BKING)
            bkpos = i;
    } while((*amp&AMARK) == 0);
    eppos = 0;
    if(amp[-1] & EPMARK)
        eppos = (amp[-1]>>5) & 077;
}

int
wattack(int pos)
{
    short *p;
    int i, j, p0, p1, b, d;

    p = attab;
    d = dir[pos];

    /* attack by knight */
    for(i=8; i--; p+=2)
        if((d & p[0]) == 0)
        if((board[pos+p[1]]&017) == WKNIGHT)
            return 0;

    /* attack along diagonal */
    i = 4;
    while(i--) {
        p0 = *p++;
        p1 = *p++;
        j = pos;
        while((dir[j]&p0) == 0) {
            j += p1;
            b = board[j];
            if(b)
            if(b == WQUEEN || b == WBISHOP)
                return 0;
            else
                break;
        }
    }

    /* attack at right angles */
    i = 4;
    while(i--) {
        p0 = *p++;
        p1 = *p++;
        j = pos;
        while((dir[j]&p0) == 0) {
            j += p1;
            b = board[j];
            if(b)
            if(b == WQUEEN || b == WROOK)
                return 0;
            else
                break;
        }
    }

    /* attack by king */
    p -= 16;
    for(i=8; i--; p+=2)
        if((d & p[0]) == 0)
        if(board[pos+p[1]] == WKING)
            return 0;

    /* attack by pawns */
    if((d & DLEFT) == 0)
    if((board[pos+7]&017) == WPAWN)
        return 0;
    if((d & DRIGHT) == 0)
    if((board[pos+9]&017) == WPAWN)
        return 0;
    return 1;
}

/*
 * database 1 and 2 ply search
 * to find best moves
 */

int
etaccess(Etab *p)
{
    uchar c[1];
    int o;

    c[0] = 0;
    o = 0;
    if(!p->error)
    switch(p->type)
    {
    case 2:
        conv2();
        c[0] = huffread(p->hp, dev.h);
        break;
    case 4:
        conv4();
        c[0] = huffread(p->hp, dev.h);
        o += 200;
        break;
    }
    if(c[0] == 0)
        return NONE;
    return c[0] + o;
}

int
etmatch(Etab *p, int *pv, int *pi, int n)
{
    int m[5], *pp, i, j, v;

    for(j=0; j<5; j++)
        m[j] = -1;
    for(i=0; i<n; i++) {
        v = pv[i];
        for(j=0; j<5; j++) {
            if(p->p[j] != v)
                continue;
            if(m[j] >= 0)
                continue;
            m[j] = pi[i];
            goto found;
        }
        return 0;
    found:;
    }
    if(m[0] < 0)
        return 0;
    dev.k1 = m[0];
    if(m[1] < 0)
        return 0;
    if(m[2] < 0) {
        v = p->p[2];
        if(v == WPAWN || v == BPAWN)
            return 0;
    }
    dev.k2 = m[1];
    dev.p1 = GONE;
    if(m[2] >= 0)
        dev.p1 = m[2];
    dev.p2 = GONE;
    if(m[3] >= 0)
        dev.p2 = m[3];
    dev.p3 = GONE;
    if(m[4] >= 0)
        dev.p3 = m[4];
    return 1;
}

int
newetab(Etab *p, int *pv, int *pi, int n)
{
    int i, ki, typ;
    char n1[10], n2[30], *s;

loop:
    p->p[0] = WKING;
    p->p[1] = BKING;
    for(ki=0; ket[ki][0]; ki++) {
        typ = ket[ki][3];
        if(typ == 0)
            continue;
        p->p[2] = ket[ki][0];
        p->p[3] = ket[ki][1];
        p->p[4] = ket[ki][2];
        if(etmatch(p, pv, pi, n))
            goto found;
    }
    badflag = 1;
    return 0;

found:
/* printf("newetab: %d %d %d %d %d\n", p->p[0], p->p[1], p->p[2], p->p[3], p->p[4]);
*/
    /*
     * build generic name W*_B*
     */
    s = n1;
    for(i=2; i<5; i++)
        if(!(p->p[i] & BLACK)) {
            *s++ = "-PNBRQK-"[p->p[i]&7];
        }
    *s++ = '_';
    for(i=2; i<5; i++)
        if(p->p[i] & BLACK) {
            *s++ = "-PNBRQK-"[p->p[i]&7];
        }
    *s = 0;
    switch(typ) {
    default:
        sprintf(n2, "%s **bad type = %d", n1, typ);
        goto bad;

    case 2:
        sprintf(n2, "E:%s.G", n1);
        p->hp = huffinit(n2);
        if(p->hp != 0)
            break;
        sprintf(n2, "F:%s.G", n1);
        p->hp = huffinit(n2);
        if(p->hp != 0)
            break;
        sprintf(n2, "G:%s.G", n1);
        p->hp = huffinit(n2);
        if(p->hp == 0)
            goto bad;
        break;

    case 4:
        sprintf(n2, "E:%s.G", n1);
        p->hp = huffinit(n2);
        if(p->hp != 0)
            break;
        sprintf(n2, "F:%s.G", n1);
        p->hp = huffinit(n2);
        if(p->hp != 0)
            break;
        sprintf(n2, "G:%s.G", n1);
        p->hp = huffinit(n2);
        if(p->hp == 0)
            goto bad;
        break;
    }
    p->type = typ;
 /*   printf(">%s/%d< ", n1, typ);   */
    badflag = 0;
    return 1;

bad:
 /*   printf("no %s ", n2); */
    ket[ki][3] = 0;
    goto loop;
}

int
gethv(void)
{
    Etab *p;
    int i, j, v;
    char b[64];
    int pv[5], pi[5];

    if(moveno & 1)
        for(i=0; i<64; i++)
            b[i] = board[i] & (BLACK|7);
    else
        for(i=0; i<64; i++) {
            v = board[63-i] & (BLACK|7);
            if(v)
                v ^= BLACK;
            b[i] = v;
        }

    /*
     * find pieces
     */
    j = 0;
    for(i=0; i<64; i++) {
        v = b[i];
        if(v == 0 || v == BLACK)
            continue;
        if(j >= 5)
            return NONE;
        pv[j] = v;
        pi[j] = i;
        j++;
    }

    /*
     * look up in table
     */
    for(p=etab; p->type; p++)
        if(etmatch(p, pv, pi, j))
            break;
    if(p->type == 0)
        if(!newetab(p, pv, pi, j))
            return NONE;
    return etaccess(p);
}

int
wgethv(void)
{
    int i, v, ov;
    short m[100];

    ov = INF;
    getmv(m);
    for(i=0; m[i]; i++) {
        move(m[i]);
        v = gethv();
        rmove();
        ov = hwcmp(ov, v);
    }
    return ov;
}

int
endgame(void)
{
    int i, j, v1, v2, v3, ov1, ov3;
    int bv1, bv2, bv3, jl, jh;
    short mv1[100], mv2[100], bm[100];

    v1 = gethv();
/*  printf("endgame: gethv() -> %ld: %d\n", dev.h, v1); */
    bv1 = INF;  /* as white to move */
    bv3 = INF;  /* as black to move */
    v3 = 0;     /* used */
    jl = 0; jh = 100;
    getmv(mv1);
    for(i=0; mv1[i]; i++) {
        ov1 = bv1;
        ov3 = bv3;

        /*
         * strong algorithm
         * one ply max weak loss
         */
        move(mv1[i]);
        v1 = gethv();
        bv1 = hscmp(bv1, v1);
        if(bv1 != ov1)
            jh = 100;
        if(v1 != NONE && v1 == bv1)
            bm[--jh] = mv1[i];

        if(bv3 != NONE) {
            /*
             * outer loop weak algorithm
             * 2 ply min
             */
            getmv(mv2);
            bv2 = INF;
            for(j=0; mv2[j]; j++) {
                /*
                 * inner loop weak algorithm
                 * 1 ply max
                 */
                move(mv2[j]);
                v2 = gethv();
                rmove();
                bv2 = hscmp(bv2, v2);
            }

            v3 = bv2;
            if(v3 == INF) {
                v3 = NONE;
                if(check())
                    v3 = 0;
                bv1 = hscmp(ov1, v3);
                v3 = hscmp(bv2, v3);
            }
            bv3 = hwcmp(bv3, v3);
            if(bv3 != ov3)
                jl = 0;
            if(v3 != NONE && v3 == bv3)
                bm[jl++] = mv1[i];
        }
        rmove();

if(pit == 0)
 {
 printf(" %2d. %-5s", i, algout(mv1[i]));
 if (badflag) badbad = 1;
 if (!(moveno%2))
  {
  if (v1 != NONE)
     printf("%3d \n", v1);
  else
     printf(" -  \n");
  }
 else
  {
  if (v3 != NONE)
     printf("%3d \n", v3);
  else
     printf(" -  \n");
  }
 }
    }
    pit++;
    if(pit == 1)
       {
       printf("\none optimal continuation:\n");
       if (moveno%2)
          printf("  1.     ...   ");
       }
    optcnt = 0;
/*  printf("w: %d", bv1);   */
    if(bv1 >= 0)
    for(i=100; i>jh; i--)
        optcnt++;

/*      printf("%s ", algout(bm[i-1]));
    printf("\n");           */

/*  printf("b: %d", bv3);   */
    if(bv3 >= 0)
    for(i=0; i<jl; i++)
        optcnt++;

/*      printf("%s ", algout(bm[i]));
    printf("\n");           */

    if(100-jh) {
        i = 99 - rand()%(100-jh);
/*      printf("bm[%d] %s\n", i, algout(bm[i]));  */
        return bm[i];
    }
    if(jl) {
        i = rand()%jl;
/*      printf("bm[%d] %s\n", i, algout(bm[i]));  */
        return bm[i];
    }
    return 0;
}

/*
 * weak max
 */
int
hwcmp(int bv, int v)
{

    if(bv == INF)
        return v;
    if(v == NONE || bv == NONE)
        return NONE;
    if(bv > v)
        return bv;
    return v;
}

/*
 * strong min
 */
int
hscmp(int bv, int v)
{
    if(bv == INF || bv == NONE)
        return v;
    if(v == NONE)
        return bv;
    if(bv < v)
        return bv;
    return v;
}

/*
 * main program
 * best play for a maximal qr.b game
 */
int
main(int argc, char *argv[])
{
    Posn p;
    int i, m;

    for(i=0; i<64; i++)
        p.bd[i] = 0;
    p.epp = 0;
    p.mvno = 0;
    p.mv[0] = 0;

    printf("\n");
    printf("Database: Ken Thompson   Interface: Herbert Pohlai\n");
    printf("Analysis of endgame position");
    for(i=1; i<argc; i++)
       {
       printf(" %s", argv[i]);
       if(argv[i][0]=='-')    /* Schwarz am Zug */
          {
          p.mvno = 1;
          break;
          }
       if(argv[i][0]=='!')    /* Kommentar */
          break;
       if((argv[i][1]<'a')||(argv[i][1]>'h')||(argv[i][2]<'1')||(argv[i][2]>'8'))
          {
          printf(" - wrong argument %s\n", argv[i]);
          exit(0);
          }
       switch(argv[i][0])
         {
         case 'P':
            p.bd[(argv[i][1]-'a') + ('8'-argv[i][2])*8] = WPAWN;
            break;          
         case 'N':          
            p.bd[(argv[i][1]-'a') + ('8'-argv[i][2])*8] = WKNIGHT;
            break;          
         case 'B':          
            p.bd[(argv[i][1]-'a') + ('8'-argv[i][2])*8] = WBISHOP;
            break;          
         case 'R':          
            p.bd[(argv[i][1]-'a') + ('8'-argv[i][2])*8] = WROOK;
            break;          
         case 'Q':          
            p.bd[(argv[i][1]-'a') + ('8'-argv[i][2])*8] = WQUEEN;
            break;          
         case 'K':          
            p.bd[(argv[i][1]-'a') + ('8'-argv[i][2])*8] = WKING;
            break;          
                            
         case 'p':          
            p.bd[(argv[i][1]-'a') + ('8'-argv[i][2])*8] = BPAWN;
            break;          
         case 'n':          
            p.bd[(argv[i][1]-'a') + ('8'-argv[i][2])*8] = BKNIGHT;
            break;
         case 'b':
            p.bd[(argv[i][1]-'a') + ('8'-argv[i][2])*8] = BBISHOP;
            break;
         case 'r':
            p.bd[(argv[i][1]-'a') + ('8'-argv[i][2])*8] = BROOK;
            break;
         case 'q':
            p.bd[(argv[i][1]-'a') + ('8'-argv[i][2])*8] = BQUEEN;
            break;
         case 'k':
            p.bd[(argv[i][1]-'a') + ('8'-argv[i][2])*8] = BKING;
            break;
         }
       }
    printf("\n\n");
    ginit(&p);

    for(;;)
      {
      restart:
        m = endgame();
        if(m == 0)      /* game over */
          {
          if (moveno < 2)
            {
            if (badbad)
               printf("\nDatabase insufficient!\n");
            else
               printf("\nWhite cannot win!\n");
            }
          break;
          }

    /*    if(m == -999)    user abort
            break;     */
/*      if(m < 0)
            {
            int count;
            for(count=m; count<0; count++)
               if (moveno>0) rmove();
            printf("Back to MOVE %d\n", (moveno+2)/2);
            goto restart;
            }                                                 */
        if (moveno%2)
           printf("(%d) %-6s\n", optcnt-1, algout(m));
        else
          {
          printf("%3d. (%d) %-6s", (moveno+2)/2, optcnt-1, algout(m));
          fflush(stdout);
          }
        move(m);
      }
    exit(0);
}

/*
 * convert ki,pi to hash
 */
void
conv2(void)
{
    int i, j, inv, t;
    long h;

    i = dev.k1;
    t = htab[i];
    inv = (t & 060) << 4;
    j = (t & 017) << 6;

    i = (t & 0300) | inv | dev.k2;
    t = gtab[i];
    j |= t & 077;
    h = (long)itab[j] << 18;

    i = (t & 0300) | inv | dev.p1;
    t = gtab[i];
    h |= (long)(t & 077) << 12;

    i = (t & 0300) | inv | dev.p2;
    t = gtab[i];
    h |= (long)(t & 077) << 6;

    i = (t & 0300) | inv | dev.p3;
    h |= (long)(gtab[i] & 077);

    dev.h = h;
}


/*
 * convert ki,pi to hash (p1 is pawn)
 */
void
conv4(void)
{
    long h;
    int p1, p2, p3, k1, k2, row;

    p1 = dev.p1;
    if(p1 != GONE) {    /* [0-5]-k1-k2-p2-p3 */
        row = p1 & 007;
        if(row >= 4) {
            p1 = prm5[p1];
            p2 = prm5[dev.p2];
            p3 = prm5[dev.p3];
            k1 = prm5[dev.k1];
            k2 = prm5[dev.k2];
            row = prm5[row];
        } else {
            p2 = dev.p2;
            p3 = dev.p3;
            k1 = dev.k1;
            k2 = dev.k2;
        }

        h = ((p1 & 070) - 010 + (row*6*8)) << 21;

        h |= p3;
        h |= p2 << 6;
        h |= k2 << 12;
        h |= k1 << 18;

        dev.h = h;
        return;
    }
bad:
    dev.h = 0;
    return;
}

/*
 * huffman decode routines
 */
void
unpackin(FILE* f, long *lp, long len, long off)
{
    long x;
    uchar *up;
    int i;

/*        printf("len: %x, off: %x\n", len, off); */

    if(sizeof(long) != 4) {
        printf("unpackin: assume long == 4 bytes\n");
        exit(1);
    }
    if((i=fread(lp, 1, len*sizeof(long), f)) != len*sizeof(long)) {
        printf("expected: %d, read: %d\n", len*sizeof(long), i);
        printf("unpackin: reading an eof\n");
        exit(1);
    }
    for(i=0; i<len; i++) {
        up = (uchar*)lp;
        x = (up[0]<<24) | (up[1]<<16) | (up[2]<<8) | up[3];
        *lp++ = x + off;
    }
}

Huff*
huffinit(char *file)
{
    Huff *h;
    long j, k, off;
    FILE* f;

    h = malloc(sizeof(Huff));
    f = fopen(file, "rb");
    if(f == NULL)
        return 0;
    h->fd = f;

    unpackin(f, &h->grouplen, 1, 0);
/* printf("grouplen=%d\n", h->grouplen); */
    unpackin(f, &h->nspec, 1, 0);
/* printf("nspec=%d\n", h->nspec);       */
    h->special = malloc(sizeof(long)*h->nspec);
    unpackin(f, h->special, h->nspec, 0);
/* printf("special=%d\n", h->special);   */

    unpackin(f, &k, 1, 0);
/* printf("k=%d\n", k);                  */
    h->tree = malloc(sizeof(long)*k);
    unpackin(f, h->tree, k, 0);
/* printf("h->tree=%d\n", h->tree);      */

    unpackin(f, &j, 1, 0);
/* printf("j=%d\n", j);                  */
    h->ztree = malloc(sizeof(long)*j);
    unpackin(f, h->ztree, j, 0);
/* printf("ztree=%d\n", h->ztree);       */

    unpackin(f, &h->indexmod, 1, 0);
/* printf("indexmod=%d\n", h->indexmod); */

    unpackin(f, &h->indexlen, 1, 0);
/* printf("indexlen=%d\n", h->indexlen); */
    h->index = malloc(sizeof(long)*h->indexlen);
    unpackin(f, h->index, h->indexlen,
        (h->indexlen + j + k + h->nspec + 6) * sizeof(long) * 8);

    return h;
}

#define BSIZE 1024
#define NEXTIN\
    if((mask>>=1) == 0) {\
        mask = 0x80;\
        if(bp >= bend) {\
            fread(b, 1, BSIZE, h->fd);\
            bp = b;\
        }\
        byte = *bp++;\
    }


#define NEXTOUT(x)\
    {\
        if(nbyte-- <= 0)\
            return x;\
    }

int
huffread(Huff *h, long byteoffset)
{
    uchar b[BSIZE], *bp, *bend;
    long code, bitoffset, nbyte, i, j, k;
    int nspeco, group[32], mask, byte;

    bitoffset = h->index[byteoffset/h->indexmod];
    nbyte = byteoffset%h->indexmod;
    nspeco = h->nspec+1;
    fseek(h->fd, bitoffset/8, 0);
    fread(b, 1, BSIZE, h->fd);
    bp = b;
    bend = b+BSIZE;
    mask = 1 << (7-(bitoffset&7));
    byte = *bp++;
    j = 0;
    for(;;) {
        if(byte & mask)
            j++;
        k = h->tree[j];
        if(k & ENDFLAG) {
            code = k & ~ENDFLAG;
            for(i=0; i<h->grouplen; i++) {
                group[i] = code%nspeco;
                code /= nspeco;
            }
            while(--i >= 0) {
                if(group[i] < h->nspec) {
                    NEXTOUT(h->special[group[i]]);
                    continue;
                }
                j = 0;
                for(;;) {
                    NEXTIN;
                    if(byte & mask)
                        j++;
                    k = h->ztree[j];
                    if(k & ENDFLAG) {
                        NEXTOUT(k & 255);
                        break;
                    }
                    j = k;
                }
            }
            k = 0;
        }
        j = k;
        NEXTIN;
    }
}

/*
 * chess algebraic output
 */
char*
algout(int m)
{
    short mv[MAXGEN];
    static char str[20];

    getamv(mv);
    xalgout(mv, m, str);
    return str;
}

void
xalgout(short *mv, int m, char *str)
{
    int p, c, t, fx, fy, tx, ty, i, j, n;

    t = (m>>12) & 017;
    c = 0;
    if(t == TENPAS || board[m&077])
        c++;
    if(t == TOO || t == TOOO) {
        *str++ = 'O';
        *str++ = '-';
        *str++ = 'O';
        if(t == TOOO) {
            *str++ = '-';
            *str++ = 'O';
        }
        goto chk;
    }
    p = board[(m>>6) & 077] & 7;
    tx = (m>>0) & 7;
    ty = (m>>3) & 7;
    if(t < TNPRO || t > TQPRO)
        t = -1;
    for(i=0; i<4; i++) {
        fx = -1;
        if((i & 1) || (c && p == 1))
            fx = (m>>6) & 7;
        fy = -1;
        if((i & 2))
            fy = (m>>9) & 7;
        n = 0;
        for(j=0; mv[j]; j++)
            if(match(mv[j], p, t, fx, fy, -1, tx, ty))
                n++;

        if(n == 1) {
            if(p != WPAWN)
                *str++ = "XPNBRQKY"[p];
            if(fx >= 0)
                *str++ = 'a'+fx;
            if(fy >= 0)
                *str++ = '8'-fy;
            if(c)
                *str++ = 'x';
            *str++ = 'a'+tx;
            *str++ = '8'-ty;
            goto chk;
        }
    }
    *str++ = '?';
    *str++ = '?';
    *str = 0;
    return;

chk:
    if(t >= TNPRO && t <= TQPRO)
        *str++ = "NBRQ"[t-TNPRO];
    move(m);
    if(check())
        if(mate())
            *str++ = '#';
        else
            *str++ = '+';
    rmove();
    *str = 0;
}

int
mate(void)
{
    short mv[MAXGEN];

    getmv(mv);
    return mv[0] == 0;
}

int
match(int m, int pf, int t, int fx, int fy, int pt, int tx, int ty)
{

    if(pf >= 0)
        if(!mat(pf, board[(m>>6)&077]&7))
            return 0;
    if(pt >= 0)
        if(!mat(pt, board[m&077]&7))
            return 0;
    if(t >= 0)
        if(((m>>12)&017) != t)
            return 0;
    if(fx >= 0)
        if(!mat(fx, (m>>6)&7))
            return 0;
    if(fy >= 0)
        if(!mat(fy, (m>>9)&7))
            return 0;
    if(tx >= 0)
        if(!mat(tx, (m>>0)&7))
            return 0;
    if(ty >= 0)
        if(!mat(ty, (m>>3)&7))
            return 0;
    return 1;
}

int
mat(int a, int b)
{

    if(a < 0)
        return 1;
    if(a >= 8) {
        if(a == 16) {
            if(b < 4)
                return 1;
            return 0;
        }
        if(a == 17) {
            if(b >= 4)
                return 1;
            return 0;
        }
        if(a-8 == b)
            return 1;
        if(15-a == b)
            return 1;
        return 0;
    }
    if(a == b)
        return 1;
    return 0;
}
